from typing import Any
from typing import List
from typing import Optional
from typing import Set
from typing import Type
from typing import Union

from pydantic import Field

from consigliere.telegram.base import Request
from consigliere.telegram.entities import ReplyMarkupType


class GetFileRequest(Request):
    file_id: str = Field(..., description="File identifier to get info about")


class GetMeRequest(Request):
    pass


class SendMessageRequest(Request):
    """
    Use this method to send text messages. On success, the sent Message is returned.
    https://core.telegram.org/bots/api#sendmessage
    """

    # fmt: off
    chat_id: Union[int, str] = Field(..., description="Unique identifier for the target chat or username of the target channel (in the format @channelusername)")
    text: str = Field(..., description="Text of the message to be sent, 1-4096 characters after entities parsing")
    parse_mode: Optional[str] = Field(None, description="Mode for parsing entities in the message text. See formatting options for more details.")
    entities: Optional[List[str]] = Field(None, description="List of special entities that appear in message text, which can be specified instead of parse_mode")
    disable_web_page_preview: Optional[bool] = Field(None, description="Disables link previews for links in this message")
    disable_notification: Optional[bool] = Field(None, description="Sends the message silently. Users will receive a notification with no sound.")
    reply_to_message_id: Optional[int] = Field(None, description="If the message is a reply, ID of the original message")
    allow_sending_without_reply: Optional[bool] = Field(None, description="Pass True, if the message should be sent even if the specified replied-to message is not found")
    reply_markup: Optional[ReplyMarkupType] = Field(None, description="Additional interface options. A JSON-serialized object for an inline keyboard, custom reply keyboard, instructions to remove reply keyboard or to force a reply from the user.")
    # fmt: on


class SendPhotoRequest(Request):
    """
    Use this method to send text messages. On success, the sent Message is returned.
    https://core.telegram.org/bots/api#sendmessage
    """

    # fmt: off
    chat_id: Union[int, str] = Field(..., description="Unique identifier for the target chat or username of the target channel (in the format @channelusername)")
    photo: Any = Field(..., description="Photo to send. Pass a file_id as String to send a photo that exists on the Telegram servers (recommended), pass an HTTP URL as a String for Telegram to get a photo from the Internet, or upload a new photo using multipart/form-data. The photo must be at most 10 MB in size. The photo's width and height must not exceed 10000 in total. Width and height ratio must be at most 20.")
    caption: Optional[str] = Field(None, description="Photo caption (may also be used when resending photos by file_id), 0-1024 characters after entities parsing")
    parse_mode: Optional[str] = Field(None, description="Mode for parsing entities in the message text. See formatting options for more details.")
    caption_entities: Optional[List[str]] = Field(None, description="A JSON-serialized list of special entities that appear in the caption, which can be specified instead of parse_mode")
    disable_notification: Optional[bool] = Field(None, description="Sends the message silently. Users will receive a notification with no sound.")
    reply_to_message_id: Optional[int] = Field(None, description="If the message is a reply, ID of the original message")
    allow_sending_without_reply: Optional[bool] = Field(None, description="Pass True, if the message should be sent even if the specified replied-to message is not found")
    reply_markup: Optional[ReplyMarkupType] = Field(None, description="Additional interface options. A JSON-serialized object for an inline keyboard, custom reply keyboard, instructions to remove reply keyboard or to force a reply from the user.")
    # fmt: on


class SetWebhookRequest(Request):
    url: str = Field(...)
    certificate: Any = Field(None)
    ip_address: Optional[str] = Field(None)
    max_connections: Optional[int] = Field(None)
    allowed_updates: Optional[List[str]] = Field(None)
    drop_pending_updates: Optional[bool] = Field(None)


class DeleteWebhookRequest(Request):
    drop_pending_updates: Optional[bool] = Field(None)


class GetWebhookInfoRequest(Request):
    pass


__models__: Set[Type[Request]] = {
    DeleteWebhookRequest,
    GetFileRequest,
    GetWebhookInfoRequest,
    SendMessageRequest,
    SendPhotoRequest,
    SetWebhookRequest,
}

__all__ = (
    "__models__",
    "DeleteWebhookRequest",
    "GetFileRequest",
    "GetWebhookInfoRequest",
    "SendMessageRequest",
    "SendPhotoRequest",
    "SetWebhookRequest",
)
